import streamlit as st

st.set_page_config(layout="wide")
st.title("Ibis benchmarking methodology")

methodology = """
See the [accompanying blog post](https://ibis-project.org/posts/ibis-bench) for more information.

## System naming convention

For convenience, we define a hyphen-separated naming convention for the 'systems' we run on:

- `ibis-*`: Ibis with various backends
    - `ibis-<backend>`: Ibis dataframe code running on the specified backend
    - `ibis-<backend>-sql`: Ibis running SQL code directly on the specified backend
- `polars-*`: Polars with various configurations
    - `polars-lazy`: Polars via `pl.LazyFrame`s
"""
st.markdown(methodology)

methodology = """
## Important considerations

Noted in more detail below (with code as the source of truth), but some important considerations include:

- TPC-H data is generated via DuckDB as Parquet files, with non-integers as decimals
- data is read in with decimals converted to floats for each query, by Ibis (via the corresponding backend) and Polars
    - additionally, hive-style partitioned columns are dropped
- queries are run 3 times and the average is taken
- all data is available in a public `gs://ibis-bench/` bucket for analysis (you can take a minimum instead of an average, for example)
- some Polars TPC-H queries were recently re-written; we will update `ibis-bench` and re-run soon including these and using newer versions of libraries
"""
st.markdown(methodology)

methodology = """
## Data generation

We use the [DuckDB TPC-H Extension](https://duckdb.org/docs/extensions/tpch.html) to generate the TPC-H data as Parquet, and optionally CSV files. The data can be partitioned.

For the current results, only Parquet data at `n=1` partitions at various scale factors are used. After generation, the data is uploaded to a cloud bucket for re-use (see the `justfile` below).

You can see the data generation source code below:
"""
st.markdown(methodology)

with open("src/ibis_bench/utils/gen_data.py") as f:
    code = f.read()

with st.expander("Show data generation code", expanded=False):
    st.code(code, line_numbers=True, language="python")

methodology = """
## Data ingestion

Data is ingested in two ways:

1. Using Ibis
2. Using Polars

All 'systems' read via Ibis **except for `polars-*`, e.g. `polars-lazy`**. 

**Important notes**:
- decimals can be converted to floats (and were for the current results)
- hive-style partitioned columns are dropped

You can see the data reading source code below:
"""
st.markdown(methodology)

with open("src/ibis_bench/utils/read_data.py") as f:
    code = f.read()

with st.expander("Show data reading code", expanded=False):
    st.code(code, line_numbers=True, language="python")

methodology = """
## Query definition

Each TPC-H query is written as a standalone Python function of the form `q<N>`, where `<N>` is the query number.

We define three sets of queries:

1. Ibis dataframe code
2. Ibis SQL code
3. Polars dataframe code

The first 10 Ibis dataframe queries were translated from the first 10 Polars queries. The remaining 12 were copied from existing Ibis TPC-H queries. The Ibis SQL queries were also directly copied from existing Ibis TPC-H SQL query code. The code was lightly modified so that functions are self-contained and for legibility when reading across systems.

**Important notes**:
- it is difficult to fairly translate SQL to dataframe code
- Polars is lacking some features (e.g. subqueries) and is in the process of re-writing some TPC-H queries
- SQL queries are transpiled to the backend's SQL dialect through Ibis (via SQLGlot)

You can see the query definition source code below:
"""
st.markdown(methodology)

with open("src/ibis_bench/queries/ibis.py") as f:
    code = f.read()

with st.expander("Show Ibis query code", expanded=False):
    st.code(code, line_numbers=True, language="python")

with open("src/ibis_bench/queries/polars.py") as f:
    code = f.read()

with st.expander("Show Polars query code", expanded=False):
    st.code(code, line_numbers=True, language="python")

with open("src/ibis_bench/queries/sql.py") as f:
    code = f.read()

with st.expander("Show Ibis SQL query code", expanded=False):
    st.code(code, line_numbers=True, language="python")

methodology = """
## Measuring query execution time

Ibis is lazyily evaluated, requiring some call by the user to trigger computation. These include methods like:

- `table.to_pandas()`
- `table.to_pyarrow()`
- `table.to_csv()`
- `table.to_parquet()`
- ...

You can also trigger computation by turning on interactive mode and outputting a table, however this may add a `LIMIT` clause to the query and thus wouldn't be a fair comparison.

Polars is eager by default, but it's generally recommended to use lazy dataframes for performance reasons. Then, you also need to call a method like:

- `.collect()`
- `.sink_parquet()`
- ...

to trigger computation. For this benchmark, we're only evaluating lazy dataframes. For consistency, **we write the results of each query to a Parquet file**. This forces computation and avoids any differences in the output.

### We are measuring...

First, we read in all 8 TPC-H tables via the corresponding lazy Parquet method. This is effectively instant, as for Ibis and Polars no data is read at this time. Then, we get a `start_time = time.time()` and run something like `q(lineitem, orders, ...).to_parquet("results/q.parquet")`. Finally, we get an `end_time = time.time()` and calculate the difference. The actual code differs (see below) but this is the general idea of what we're measuring and how.
"""
st.markdown(methodology)

with open("src/ibis_bench/utils/write_data.py") as f:
    code = f.read()

with st.expander("Show data writing code", expanded=False):
    st.code(code, line_numbers=True, language="python")

with open("src/ibis_bench/utils/monitor.py") as f:
    code = f.read()

with st.expander("Show monitoring code", expanded=False):
    st.code(code, line_numbers=True, language="python")

methodology = """
## Running the benchmark

To run the benchmark, I wanted a CLI helper tool and ended up making two. The first is used to generate the TPC-H data and run the queries. The second mainly functions as a wrapper around the first to run it in a subprocess. This is because queries can and do fail by being killed by the operating system, generally when memory pressure is too high. To avoid stopping the benchmark, we run each query in a subprocess and continue on failures.

The second CLI also has a tool for combining the JSON timings data for each query into a single Parquet file. This is useful because reading thousands of files from a cloud bucket is very slow.

Currently, each query is run 3 times. In visualizing, the average is taken.

### The `bench` CLI tool

The `bench` tool has a `gen-data` command and a `run` command that work across different systems, scale factors, partitions, and more.
"""
st.markdown(methodology)

with open("src/ibis_bench/cli.py") as f:
    code = f.read()

with st.expander("Show `bench` code", expanded=False):
    st.code(code, line_numbers=True, language="python")

methodology = """
### The `bench2` CLI tool

The `bench2` tool has a `combine-json` command and a `run` command.

"""
st.markdown(methodology)

with open("src/ibis_bench/cli2.py") as f:
    code = f.read()

with st.expander("Show `bench2` code", expanded=False):
    st.code(code, line_numbers=True, language="python")

methodology = """
## Running on the cloud

Most runs are done on GCP VMs. The `justfile` in the repo contains common commands that are used throughout, especially for cloud things. An outline to repeat the process follows.

Edit the `justfile` to the desired VM settings (name, type, region).

Create the VM:

```bash
just create-vm
```

SSH into the VM:

```bash
just vm-ssh
```

Clone and change into the `ibis-bench` repo:

```bash
git clone https://github.com/lostmygithubaccount/ibis-bench
cd ibis-bench
```

Run the VM bootstrap script:

```bash
bash vm-bootstrap.sh
```

This:

- installs `just`
- installs `python3.11`
- installs and sets up some other useful things

Source the `~/.bashrc` file for changes to take effect:

```bash
source ~/.bashrc
```

Now install `ibis-bench`:

```bash
just setup
```

Download the TPC-H data:

```bash
just tpch-download-parquet
```

And run the benchmark:

```bash
just run-all-parquet
```

This will run the process in the background and continue running even if the SSH connection is lost. Logs are written to `out.log` and you can also use `top` to check that `bench` is running.

Once the benchmark is complete, combine the JSON files:

```bash
bench2 combine-json "<INSTANCE_NAME>"
```

This is done manually to avoid mistakes with the `justfile`.

Then authenticate with GCP so we can upload the logs to the bucket:

```bash
gcloud auth login
```

Upload the logs:

```bash
just logs-upload
```

Exit from the VM and delete it:
    
```bash
exit
just vm-delete
```

And that's it! The logs are now in the bucket and can be analyzed.
"""
st.markdown(methodology)


with open("justfile") as f:
    code = f.read()

with st.expander("Show Justfile", expanded=False):
    st.code(code, line_numbers=True, language="justfile")

with open("vm-bootstrap.sh") as f:
    code = f.read()

with st.expander("Show VM bootstrap script", expanded=False):
    st.code(code, line_numbers=True, language="bash")

methodology = """
## Results

Results are publicly available in the `gs://ibis-bench/` bucket. The `bench_logs_v2` directory contains current results.
"""
st.markdown(methodology)



import ibis
import polars as pl
import ibis.selectors as s
import polars.selectors as ps

from ibis_bench.utils.gen_data import get_data_dir

DECIMAL_TO_FLOAT = True


def get_ibis_tables(
    sf, n_partitions=1, con=ibis.connect("duckdb://"), csv=False, decimal_to_float=True
):
    data_directory = get_data_dir(sf, n_partitions, csv=csv)

    if not csv:
        customer = con.read_parquet(
            f"{data_directory}/customer/*.parquet", table_name="customer"
        )
        lineitem = con.read_parquet(
            f"{data_directory}/lineitem/*.parquet", table_name="lineitem"
        )
        nation = con.read_parquet(
            f"{data_directory}/nation/*.parquet", table_name="nation"
        )
        orders = con.read_parquet(
            f"{data_directory}/orders/*.parquet", table_name="orders"
        )
        part = con.read_parquet(f"{data_directory}/part/*.parquet", table_name="part")
        partsupp = con.read_parquet(
            f"{data_directory}/partsupp/*.parquet", table_name="partsupp"
        )
        region = con.read_parquet(
            f"{data_directory}/region/*.parquet", table_name="region"
        )
        supplier = con.read_parquet(
            f"{data_directory}/supplier/*.parquet", table_name="supplier"
        )
    else:
        customer = con.read_csv(
            f"{data_directory}/customer/*.csv", table_name="customer"
        )
        lineitem = con.read_csv(
            f"{data_directory}/lineitem/*.csv", table_name="lineitem"
        )
        nation = con.read_csv(f"{data_directory}/nation/*.csv", table_name="nation")
        orders = con.read_csv(f"{data_directory}/orders/*.csv", table_name="orders")
        part = con.read_csv(f"{data_directory}/part/*.csv", table_name="part")
        partsupp = con.read_csv(
            f"{data_directory}/partsupp/*.csv", table_name="partsupp"
        )
        region = con.read_csv(f"{data_directory}/region/*.csv", table_name="region")
        supplier = con.read_csv(
            f"{data_directory}/supplier/*.csv", table_name="supplier"
        )

    # TODO: report issue(s) (DataFusion backend issue)
    def _decimal_to_float(t):
        return t.mutate(
            s.across(
                s.of_type("decimal"),
                ibis._.cast("float"),
            )
        )

    if decimal_to_float:
        customer = _decimal_to_float(customer)
        lineitem = _decimal_to_float(lineitem)
        nation = _decimal_to_float(nation)
        orders = _decimal_to_float(orders)
        part = _decimal_to_float(part)
        partsupp = _decimal_to_float(partsupp)
        region = _decimal_to_float(region)
        supplier = _decimal_to_float(supplier)

    # TODO: keep this or figure something out and remove
    def _drop_hive_cols(t):
        # NOTE: some backends don't create the hive-partitioned columns (at least by default)
        # DuckDB and Polars do, DataFusion doesn't, so first check if the column(s) exist
        if "sf" in t.columns:
            t = t.drop("sf")
        if "n" in t.columns:
            t = t.drop("n")
        return t

    customer = _drop_hive_cols(customer)
    lineitem = _drop_hive_cols(lineitem)
    nation = _drop_hive_cols(nation)
    orders = _drop_hive_cols(orders)
    part = _drop_hive_cols(part)
    partsupp = _drop_hive_cols(partsupp)
    region = _drop_hive_cols(region)
    supplier = _drop_hive_cols(supplier)

    return customer, lineitem, nation, orders, part, partsupp, region, supplier


def get_polars_tables(
    sf, n_partitions=1, lazy=True, csv=False, decimal_to_float=DECIMAL_TO_FLOAT
):
    data_directory = get_data_dir(sf, n_partitions, csv=csv)

    if not csv:
        if lazy:
            customer = pl.scan_parquet(f"{data_directory}/customer/*.parquet")
            lineitem = pl.scan_parquet(f"{data_directory}/lineitem/*.parquet")
            nation = pl.scan_parquet(f"{data_directory}/nation/*.parquet")
            orders = pl.scan_parquet(f"{data_directory}/orders/*.parquet")
            part = pl.scan_parquet(f"{data_directory}/part/*.parquet")
            partsupp = pl.scan_parquet(f"{data_directory}/partsupp/*.parquet")
            region = pl.scan_parquet(f"{data_directory}/region/*.parquet")
            supplier = pl.scan_parquet(f"{data_directory}/supplier/*.parquet")
        else:
            customer = pl.read_parquet(f"{data_directory}/customer/*.parquet")
            lineitem = pl.read_parquet(f"{data_directory}/lineitem/*.parquet")
            nation = pl.read_parquet(f"{data_directory}/nation/*.parquet")
            orders = pl.read_parquet(f"{data_directory}/orders/*.parquet")
            part = pl.read_parquet(f"{data_directory}/part/*.parquet")
            partsupp = pl.read_parquet(f"{data_directory}/partsupp/*.parquet")
            region = pl.read_parquet(f"{data_directory}/region/*.parquet")
            supplier = pl.read_parquet(f"{data_directory}/supplier/*.parquet")
    else:
        if lazy:
            customer = pl.scan_csv(f"{data_directory}/customer/*.csv")
            lineitem = pl.scan_csv(f"{data_directory}/lineitem/*.csv")
            nation = pl.scan_csv(f"{data_directory}/nation/*.csv")
            orders = pl.scan_csv(f"{data_directory}/orders/*.csv")
            part = pl.scan_csv(f"{data_directory}/part/*.csv")
            partsupp = pl.scan_csv(f"{data_directory}/partsupp/*.csv")
            region = pl.scan_csv(f"{data_directory}/region/*.csv")
            supplier = pl.scan_csv(f"{data_directory}/supplier/*.csv")
        else:
            customer = pl.read_csv(f"{data_directory}/customer/*.csv")
            lineitem = pl.read_csv(f"{data_directory}/lineitem/*.csv")
            nation = pl.read_csv(f"{data_directory}/nation/*.csv")
            orders = pl.read_csv(f"{data_directory}/orders/*.csv")
            part = pl.read_csv(f"{data_directory}/part/*.csv")
            partsupp = pl.read_csv(f"{data_directory}/partsupp/*.csv")
            region = pl.read_csv(f"{data_directory}/region/*.csv")
            supplier = pl.read_csv(f"{data_directory}/supplier/*.csv")

    # TODO: report issue(s) (issue(s) at higher SFs)
    def _decimal_to_float(df):
        return df.with_columns((ps.decimal().cast(pl.Float64)))

    if decimal_to_float:
        customer = _decimal_to_float(customer)
        lineitem = _decimal_to_float(lineitem)
        nation = _decimal_to_float(nation)
        orders = _decimal_to_float(orders)
        part = _decimal_to_float(part)
        partsupp = _decimal_to_float(partsupp)
        region = _decimal_to_float(region)
        supplier = _decimal_to_float(supplier)

    # TODO: keep this or figure something out and remove
    # def _drop_hive_cols(df):
    #     return df.drop(["sf", "n"])

    # customer = _drop_hive_cols(customer)
    # lineitem = _drop_hive_cols(lineitem)
    # nation = _drop_hive_cols(nation)
    # orders = _drop_hive_cols(orders)
    # part = _drop_hive_cols(part)
    # partsupp = _drop_hive_cols(partsupp)
    # region = _drop_hive_cols(region)
    # supplier = _drop_hive_cols(supplier)

    return customer, lineitem, nation, orders, part, partsupp, region, supplier



import ibis
import polars as pl
import ibis.selectors as s
import polars.selectors as ps

from ibis_bench.tpcds.utils.gen_data import get_data_dir


def get_ibis_tables(
    sf, con=ibis.connect("duckdb://"), csv=False, decimal_to_float=False
):
    data_directory = get_data_dir(sf, csv=csv)

    if not csv:
        call_center = con.read_parquet(
            f"{data_directory}/call_center/*.parquet", table_name="call_center"
        )
        catalog_page = con.read_parquet(
            f"{data_directory}/catalog_page/*.parquet", table_name="catalog_page"
        )
        catalog_returns = con.read_parquet(
            f"{data_directory}/catalog_returns/*.parquet", table_name="catalog_returns"
        )
        catalog_sales = con.read_parquet(
            f"{data_directory}/catalog_sales/*.parquet", table_name="catalog_sales"
        )
        customer = con.read_parquet(
            f"{data_directory}/customer/*.parquet", table_name="customer"
        )
        customer_address = con.read_parquet(
            f"{data_directory}/customer_address/*.parquet",
            table_name="customer_address",
        )
        customer_demographics = con.read_parquet(
            f"{data_directory}/customer_demographics/*.parquet",
            table_name="customer_demographics",
        )
        date_dim = con.read_parquet(
            f"{data_directory}/date_dim/*.parquet", table_name="date_dim"
        )
        household_demographics = con.read_parquet(
            f"{data_directory}/household_demographics/*.parquet",
            table_name="household_demographics",
        )
        income_band = con.read_parquet(
            f"{data_directory}/income_band/*.parquet", table_name="income_band"
        )
        inventory = con.read_parquet(
            f"{data_directory}/inventory/*.parquet", table_name="inventory"
        )
        item = con.read_parquet(f"{data_directory}/item/*.parquet", table_name="item")
        promotion = con.read_parquet(
            f"{data_directory}/promotion/*.parquet", table_name="promotion"
        )
        reason = con.read_parquet(
            f"{data_directory}/reason/*.parquet", table_name="reason"
        )
        ship_mode = con.read_parquet(
            f"{data_directory}/ship_mode/*.parquet", table_name="ship_mode"
        )
        store = con.read_parquet(
            f"{data_directory}/store/*.parquet", table_name="store"
        )
        store_returns = con.read_parquet(
            f"{data_directory}/store_returns/*.parquet", table_name="store_returns"
        )
        store_sales = con.read_parquet(
            f"{data_directory}/store_sales/*.parquet", table_name="store_sales"
        )
        time_dim = con.read_parquet(
            f"{data_directory}/time_dim/*.parquet", table_name="time_dim"
        )
        warehouse = con.read_parquet(
            f"{data_directory}/warehouse/*.parquet", table_name="warehouse"
        )
        web_page = con.read_parquet(
            f"{data_directory}/web_page/*.parquet", table_name="web_page"
        )
        web_returns = con.read_parquet(
            f"{data_directory}/web_returns/*.parquet", table_name="web_returns"
        )
        web_sales = con.read_parquet(
            f"{data_directory}/web_sales/*.parquet", table_name="web_sales"
        )
        web_site = con.read_parquet(
            f"{data_directory}/web_site/*.parquet", table_name="web_site"
        )
    else:
        call_center == con.read_csv(
            f"{data_directory}/call_center/*.csv", table_name="call_center"
        )
        catalog_page = con.read_csv(
            f"{data_directory}/catalog_page/*.csv", table_name="catalog_page"
        )
        catalog_returns = con.read_csv(
            f"{data_directory}/catalog_returns/*.csv", table_name="catalog_returns"
        )
        catalog_sales = con.read_csv(
            f"{data_directory}/catalog_sales/*.csv", table_name="catalog_sales"
        )
        customer = con.read_csv(
            f"{data_directory}/customer/*.csv", table_name="customer"
        )
        customer_address = con.read_csv(
            f"{data_directory}/customer_address/*.csv", table_name="customer_address"
        )
        customer_demographics = con.read_csv(
            f"{data_directory}/customer_demographics/*.csv",
            table_name="customer_demographics",
        )
        date_dim = con.read_csv(
            f"{data_directory}/date_dim/*.csv", table_name="date_dim"
        )
        household_demographics = con.read_csv(
            f"{data_directory}/household_demographics/*.csv",
            table_name="household_demographics",
        )
        income_band = con.read_csv(
            f"{data_directory}/income_band/*.csv", table_name="income_band"
        )
        inventory = con.read_csv(
            f"{data_directory}/inventory/*.csv", table_name="inventory"
        )
        item = con.read_csv(f"{data_directory}/item/*.csv", table_name="item")
        promotion = con.read_csv(
            f"{data_directory}/promotion/*.csv", table_name="promotion"
        )
        reason = con.read_csv(f"{data_directory}/reason/*.csv", table_name="reason")
        ship_mode = con.read_csv(
            f"{data_directory}/ship_mode/*.csv", table_name="ship_mode"
        )
        store = con.read_csv(f"{data_directory}/store/*.csv", table_name="store")
        store_returns = con.read_csv(
            f"{data_directory}/store_returns/*.csv", table_name="store_returns"
        )
        store_sales = con.read_csv(
            f"{data_directory}/store_sales/*.csv", table_name="store_sales"
        )
        time_dim = con.read_csv(
            f"{data_directory}/time_dim/*.csv", table_name="time_dim"
        )
        warehouse = con.read_csv(
            f"{data_directory}/warehouse/*.csv", table_name="warehouse"
        )
        web_page = con.read_csv(
            f"{data_directory}/web_page/*.csv", table_name="web_page"
        )
        web_returns = con.read_csv(
            f"{data_directory}/web_returns/*.csv", table_name="web_returns"
        )
        web_sales = con.read_csv(
            f"{data_directory}/web_sales/*.csv", table_name="web_sales"
        )
        web_site = con.read_csv(
            f"{data_directory}/web_site/*.csv", table_name="web_site"
        )

    all_tables = (
        call_center,
        catalog_page,
        catalog_returns,
        catalog_sales,
        customer,
        customer_address,
        customer_demographics,
        date_dim,
        household_demographics,
        income_band,
        inventory,
        item,
        promotion,
        reason,
        ship_mode,
        store,
        store_returns,
        store_sales,
        time_dim,
        warehouse,
        web_page,
        web_returns,
        web_sales,
        web_site,
    )

    # TODO: report issue(s) (DataFusion backend issue)
    def _decimal_to_float(t):
        return t.mutate(
            s.across(
                s.of_type("decimal"),
                ibis._.cast("float"),
            )
        )

    if decimal_to_float:
        all_tables = tuple(map(_decimal_to_float, all_tables))

    # TODO: keep this or figure something out and remove
    def _drop_hive_cols(t):
        # NOTE: some backends don't create the hive-partitioned columns (at least by default)
        # DuckDB and Polars do, DataFusion doesn't, so first check if the column(s) exist
        if "sf" in t.columns:
            t = t.drop("sf")
        if "n" in t.columns:
            t = t.drop("n")
        return t

    all_tables = tuple(map(_drop_hive_cols, all_tables))

    return all_tables



import sys
import streamlit as st

sys.path.append("..")

from results import get_t, get_instance_details

# st.set_page_config(layout="wide")
st.title("Ibis benchmarking instance type details")

t = get_t()
instance_details = get_instance_details(t)

# st.dataframe(instance_details, use_container_width=True)
st.table(instance_details)



# imports
import ibis
import uuid
import typer

from ibis_bench.utils.logging import log
from ibis_bench.utils.monitor import monitor_it

## import tpch functions
from ibis_bench.tpch.utils.gen_data import generate_data as tpch_generate_data
from ibis_bench.tpch.utils.read_data import (
    get_ibis_tables as tpch_get_ibis_tables,
    get_polars_tables as tpch_get_polars_tables,
)

## import tpcds functions
from ibis_bench.tpcds.utils.gen_data import generate_data as tpcds_generate_data
# from ibis_bench.tpcds.utils.read_data import (
#     get_ibis_tables as tpcds_get_ibis_tables,
#     get_polars_tables as tpcds_get_polars_tables,
# )

# defaults
DEFAULT_SCALE_FACTORS = [1]
DEFAULT_N_PARTITIONS = [1]

TYPER_KWARGS = {
    "no_args_is_help": True,
    "add_completion": False,
    "context_settings": {"help_option_names": ["-h", "--help"]},
}

# typer apps
app = typer.Typer(help="ibis-bench", **TYPER_KWARGS)
tpch_app = typer.Typer(help="TPC-H commands", **TYPER_KWARGS)
tpcds_app = typer.Typer(help="TPC-DS commands", **TYPER_KWARGS)

## add subcommands
app.add_typer(tpch_app, name="tpch")
app.add_typer(tpcds_app, name="tpcds")

## add subcommand aliases
app.add_typer(tpch_app, name="h", hidden=True)
app.add_typer(tpcds_app, name="ds", hidden=True)


# TPC-H commands
@tpch_app.command("gen")
def tpch_gen_data(
    scale_factors: list[int] = typer.Option(
        DEFAULT_SCALE_FACTORS, "--scale-factor", "-s", help="scale factors"
    ),
    n_partitions: list[int] = typer.Option(
        DEFAULT_N_PARTITIONS, "--n-partitions", "-n", help="number of partitions"
    ),
    csv: bool = typer.Option(
        False, "--csv", "-c", help="generate CSV files in addition to Parquet"
    ),
):
    """
    generate tpc-h benchmarking data
    """
    for sf in sorted(scale_factors):
        for n in sorted(n_partitions):
            tpch_generate_data(sf, n, csv=csv)


@tpch_app.command("run")
def tpch_run(
    systems: list[str] = typer.Argument(..., help="system to run on"),
    scale_factors: list[int] = typer.Option(
        DEFAULT_SCALE_FACTORS, "--scale-factor", "-s", help="scale factors"
    ),
    n_partitions: list[int] = typer.Option(
        DEFAULT_N_PARTITIONS, "--n-partitions", "-n", help="number of partitions"
    ),
    q_number: list[int] = typer.Option(
        None, "--query-number", "-q", help="query numbers"
    ),
    exclude_queries: list[int] = typer.Option(
        None, "--exclude-queries", "-e", help="exclude query numbers"
    ),
    use_csv: bool = typer.Option(
        False, "--csv", "-c", help="use CSV files instead of parquet"
    ),
    decimal_to_float: bool = typer.Option(
        False, "--decimal-to-float", "-d", help="convert decimal to float"
    ),
    instance_type: str = typer.Option(
        None, "--instance-type", "-i", help="instance type"
    ),
):
    """
    run tpc-h benchmarking queries
    """
    session_id = str(uuid.uuid4())

    for sf in sorted(scale_factors):
        for n in sorted(n_partitions):
            for system in systems:
                system_parts = system.split("-")

                if system_parts[0] == "ibis" and system_parts[-1] != "sql":
                    backend = system_parts[1]
                    con = ibis.connect(f"{backend}://")
                    log.info(f"connected to {backend} backend")

                    from ibis_bench.tpch.queries.ibis import all_queries

                    (
                        customer,
                        lineitem,
                        nation,
                        orders,
                        part,
                        partsupp,
                        region,
                        supplier,
                    ) = tpch_get_ibis_tables(
                        sf=sf,
                        n_partitions=n,
                        con=con,
                        csv=use_csv,
                        decimal_to_float=decimal_to_float,
                    )
                elif system_parts[0] == "ibis" and system_parts[-1] == "sql":
                    backend = system_parts[1]
                    con = ibis.connect(f"{backend}://")
                    log.info(f"connected to {backend} backend (for SQL)")

                    from ibis_bench.tpch.queries.sql import all_queries

                    (
                        customer,
                        lineitem,
                        nation,
                        orders,
                        part,
                        partsupp,
                        region,
                        supplier,
                    ) = tpch_get_ibis_tables(
                        sf=sf,
                        n_partitions=n,
                        con=con,
                        csv=use_csv,
                        decimal_to_float=decimal_to_float,
                    )
                elif system_parts[0] == "polars":
                    backend = system_parts[0]
                    lazy = system_parts[1] == "lazy"
                    log.info(f"using Polars with lazy={lazy}")

                    from ibis_bench.tpch.queries.polars import all_queries

                    (
                        customer,
                        lineitem,
                        nation,
                        orders,
                        part,
                        partsupp,
                        region,
                        supplier,
                    ) = tpch_get_polars_tables(
                        sf=sf,
                        n_partitions=n,
                        lazy=lazy,
                        csv=use_csv,
                        decimal_to_float=decimal_to_float,
                    )

                queries = {
                    q: query
                    for q, query in enumerate(all_queries, start=1)
                    if (q_number is None or q in q_number)
                    and (exclude_queries is None or q not in exclude_queries)
                }

                for q, query in queries.items():
                    try:
                        monitor_it(
                            query,
                            sf=sf,
                            n_partitions=n,
                            query_number=q,
                            system=system,
                            session_id=session_id,
                            instance_type=instance_type,
                            use_csv=use_csv,
                            decimal_to_float=decimal_to_float,
                            # tpch tables
                            customer=customer,
                            lineitem=lineitem,
                            nation=nation,
                            orders=orders,
                            part=part,
                            partsupp=partsupp,
                            region=region,
                            supplier=supplier,
                            # used for SQL runs
                            dialect=backend if backend else None,
                        )
                    except Exception as e:
                        log.info(
                            f"error running query {q} at scale factor {sf} and {n} partitions: {e}"
                        )


# TPC-DS commands
@tpcds_app.command("gen")
def tpcds_gen_data(
    scale_factors: list[int] = typer.Option(
        DEFAULT_SCALE_FACTORS, "--scale-factor", "-s", help="scale factors"
    ),
    csv: bool = typer.Option(
        False, "--csv", "-c", help="generate CSV files in addition to Parquet"
    ),
):
    """
    generate tpc-h benchmarking data
    """
    for sf in sorted(scale_factors):
        tpcds_generate_data(sf, csv=csv)


if __name__ == "__main__":
    typer.run(app)



# adapted from corresponding Polars queries (first ten)
# also adapted from: https://github.com/ibis-project/ibis/tree/main/ibis/backends/tests/tpch

import ibis

from datetime import date


def q1(lineitem, **kwargs):
    var1 = date(1998, 9, 2)

    q_final = (
        lineitem.filter(lineitem["l_shipdate"] <= var1)
        .group_by(["l_returnflag", "l_linestatus"])
        .aggregate(
            sum_qty=lineitem["l_quantity"].sum(),
            sum_base_price=lineitem["l_extendedprice"].sum(),
            sum_disc_price=(
                lineitem["l_extendedprice"] * (1 - lineitem["l_discount"])
            ).sum(),
            sum_charge=(
                lineitem["l_extendedprice"]
                * (1 - lineitem["l_discount"])
                * (1 + lineitem["l_tax"])
            ).sum(),
            avg_qty=lineitem["l_quantity"].mean(),
            avg_price=lineitem["l_extendedprice"].mean(),
            avg_disc=lineitem["l_discount"].mean(),
            count_order=lambda lineitem: lineitem.count(),
        )
        .order_by(["l_returnflag", "l_linestatus"])
    )
    return q_final


def q2(nation, part, partsupp, region, supplier, **kwargs):
    var1 = 15
    var2 = "BRASS"
    var3 = "EUROPE"

    q1 = (
        part.join(partsupp, part["p_partkey"] == partsupp["ps_partkey"])
        .join(supplier, partsupp["ps_suppkey"] == supplier["s_suppkey"])
        .join(nation, supplier["s_nationkey"] == nation["n_nationkey"])
        .join(region, nation["n_regionkey"] == region["r_regionkey"])
        .filter(ibis._["p_size"] == var1)
        .filter(ibis._["p_type"].endswith(var2))
        .filter(ibis._["r_name"] == var3)
    )

    q_final = (
        q1.group_by("p_partkey")
        .agg(ps_supplycost=ibis._["ps_supplycost"].min())
        .join(q1, ["p_partkey"])
        .select(
            "s_acctbal",
            "s_name",
            "n_name",
            "p_partkey",
            "p_mfgr",
            "s_address",
            "s_phone",
            "s_comment",
        )
        .order_by(ibis.desc("s_acctbal"), "n_name", "s_name", "p_partkey")
        .limit(100)
    )

    return q_final


def q3(customer, lineitem, orders, **kwargs):
    var1 = "BUILDING"
    var2 = date(1995, 3, 15)

    q_final = (
        customer.filter(customer["c_mktsegment"] == var1)
        .join(orders, customer["c_custkey"] == orders["o_custkey"])
        .join(lineitem, orders["o_orderkey"] == lineitem["l_orderkey"])
        .filter(ibis._["o_orderdate"] < var2)
        .filter(ibis._["l_shipdate"] > var2)
        .mutate(revenue=(lineitem["l_extendedprice"] * (1 - lineitem["l_discount"])))
        .group_by(
            "o_orderkey",
            "o_orderdate",
            "o_shippriority",
        )
        .agg(revenue=ibis._["revenue"].sum())
        .select(
            ibis._["o_orderkey"].name("o_orderkey"),
            "revenue",
            "o_orderdate",
            "o_shippriority",
        )
        .order_by(ibis.desc("revenue"), "o_orderdate")
        .limit(10)
    )

    return q_final


def q4(lineitem, orders, **kwargs):
    var1 = date(1993, 7, 1)
    var2 = date(1993, 10, 1)

    q_final = (
        lineitem.join(orders, lineitem["l_orderkey"] == orders["o_orderkey"])
        .filter((orders["o_orderdate"] >= var1) & (orders["o_orderdate"] < var2))
        .filter(lineitem["l_commitdate"] < lineitem["l_receiptdate"])
        .distinct(on=["o_orderpriority", "l_orderkey"])
        .group_by("o_orderpriority")
        .agg(order_count=ibis._.count())
        .order_by("o_orderpriority")
    )

    return q_final


def q5(region, nation, supplier, customer, orders, lineitem, **kwargs):
    var1 = "ASIA"
    var2 = date(1994, 1, 1)
    var3 = date(1995, 1, 1)

    q_final = (
        region.join(nation, region["r_regionkey"] == nation["n_regionkey"])
        .join(customer, ibis._["n_nationkey"] == customer["c_nationkey"])
        .join(orders, ibis._["c_custkey"] == orders["o_custkey"])
        .join(lineitem, ibis._["o_orderkey"] == lineitem["l_orderkey"])
        .join(
            supplier,
            (ibis._["l_suppkey"] == supplier["s_suppkey"])
            & (ibis._["n_nationkey"] == supplier["s_nationkey"]),
        )
        .filter(ibis._["r_name"] == var1)
        .filter((ibis._["o_orderdate"] >= var2) & (ibis._["o_orderdate"] < var3))
        .mutate(revenue=(lineitem["l_extendedprice"] * (1 - lineitem["l_discount"])))
        .group_by("n_name")
        .agg(revenue=ibis._["revenue"].sum())
        .order_by(ibis.desc("revenue"))
    )

    return q_final


def q6(lineitem, **kwargs):
    var1 = date(1994, 1, 1)
    var2 = date(1995, 1, 1)
    var3 = 0.05
    var4 = 0.07
    var5 = 24

    q_final = (
        lineitem.filter(
            (lineitem["l_shipdate"] >= var1) & (lineitem["l_shipdate"] < var2)
        )
        .filter((lineitem["l_discount"] >= var3) & (lineitem["l_discount"] <= var4))
        .filter(lineitem["l_quantity"] < var5)
        .mutate(revenue=ibis._["l_extendedprice"] * (ibis._["l_discount"]))
        .agg(revenue=ibis._["revenue"].sum())
    )

    return q_final


def q7(customer, lineitem, nation, orders, supplier, **kwargs):
    var1 = "FRANCE"
    var2 = "GERMANY"
    var3 = date(1995, 1, 1)
    var4 = date(1996, 12, 31)

    n1 = nation.filter(nation["n_name"] == var1)
    n2 = nation.filter(nation["n_name"] == var2)

    q1 = (
        customer.join(n1, customer["c_nationkey"] == n1["n_nationkey"])
        .join(orders, customer["c_custkey"] == orders["o_custkey"])
        .rename({"cust_nation": "n_name"})
        .join(lineitem, orders["o_orderkey"] == lineitem["l_orderkey"])
        .join(supplier, lineitem["l_suppkey"] == supplier["s_suppkey"])
        .join(n2, supplier["s_nationkey"] == n2["n_nationkey"])
        .rename({"supp_nation": "n_name"})
    )

    q2 = (
        customer.join(n2, customer["c_nationkey"] == n2["n_nationkey"])
        .join(orders, customer["c_custkey"] == orders["o_custkey"])
        .rename({"cust_nation": "n_name"})
        .join(lineitem, orders["o_orderkey"] == lineitem["l_orderkey"])
        .join(supplier, lineitem["l_suppkey"] == supplier["s_suppkey"])
        .join(n1, supplier["s_nationkey"] == n1["n_nationkey"])
        .rename({"supp_nation": "n_name"})
    )

    q_final = (
        q1.union(q2)
        .filter((ibis._["l_shipdate"] >= var3) & (ibis._["l_shipdate"] <= var4))
        .mutate(
            volume=(ibis._["l_extendedprice"] * (1 - ibis._["l_discount"])),
            l_year=ibis._["l_shipdate"].year(),
        )
        .group_by("supp_nation", "cust_nation", "l_year")
        .agg(revenue=ibis._["volume"].sum())
        .order_by("supp_nation", "cust_nation", "l_year")
    )

    return q_final


def q8(customer, lineitem, nation, orders, part, region, supplier, **kwargs):
    var1 = "BRAZIL"
    var2 = "AMERICA"
    var3 = "ECONOMY ANODIZED STEEL"
    var4 = date(1995, 1, 1)
    var5 = date(1996, 12, 31)

    n1 = nation.select("n_nationkey", "n_regionkey")
    n2 = nation.select("n_nationkey", "n_name")

    q_final = (
        part.join(lineitem, lineitem["l_partkey"] == part["p_partkey"])
        .join(supplier, lineitem["l_suppkey"] == supplier["s_suppkey"])
        .join(orders, lineitem["l_orderkey"] == orders["o_orderkey"])
        .join(customer, orders["o_custkey"] == customer["c_custkey"])
        .join(n1, customer["c_nationkey"] == n1["n_nationkey"])
        .join(region, n1["n_regionkey"] == region["r_regionkey"])
        .filter(region["r_name"] == var2)
        .join(n2, supplier["s_nationkey"] == n2["n_nationkey"])
        .filter((orders["o_orderdate"] >= var4) & (orders["o_orderdate"] <= var5))
        .filter(part["p_type"] == var3)
        .select(
            o_year=orders["o_orderdate"].year(),
            volume=lineitem["l_extendedprice"] * (1 - lineitem["l_discount"]),
            nation=n2["n_name"],
        )
        .mutate(
            _tmp=ibis.case()
            .when(
                ibis._["nation"] == var1,
                ibis._["volume"],
            )
            .else_(0)
            .end()
        )
        .group_by("o_year")
        .agg(mkt_share=(ibis._["_tmp"].sum() / ibis._["volume"].sum()).round(2))
        .order_by("o_year")
    )

    return q_final


def q9(lineitem, nation, orders, part, partsupp, supplier, **kwargs):
    q_final = (
        lineitem.join(supplier, lineitem["l_suppkey"] == supplier["s_suppkey"])
        .join(
            partsupp,
            (lineitem["l_suppkey"] == partsupp["ps_suppkey"])
            & (lineitem["l_partkey"] == partsupp["ps_partkey"]),
        )
        .join(part, lineitem["l_partkey"] == part["p_partkey"])
        .join(orders, lineitem["l_orderkey"] == orders["o_orderkey"])
        .join(nation, supplier["s_nationkey"] == nation["n_nationkey"])
        .filter(ibis._["p_name"].contains("green"))
        .select(
            nation=ibis._["n_name"],
            o_year=ibis._["o_orderdate"].year(),
            amount=ibis._["l_extendedprice"] * (1 - ibis._["l_discount"])
            - ibis._["ps_supplycost"] * ibis._["l_quantity"],
        )
        .group_by("nation", "o_year")
        .agg(sum_profit=ibis._["amount"].sum().round(2))
        .order_by("nation", ibis.desc("o_year"))
    )

    return q_final


def q10(customer, lineitem, nation, orders, **kwargs):
    var1 = date(1993, 10, 1)
    var2 = date(1994, 1, 1)

    q_final = (
        customer.join(orders, customer["c_custkey"] == orders["o_custkey"])
        .join(lineitem, orders["o_orderkey"] == lineitem["l_orderkey"])
        .join(nation, customer["c_nationkey"] == nation["n_nationkey"])
        .filter((ibis._["o_orderdate"] >= var1) & (ibis._["o_orderdate"] < var2))
        .filter(ibis._["l_returnflag"] == "R")
        .group_by(
            "c_custkey",
            "c_name",
            "c_acctbal",
            "c_phone",
            "n_name",
            "c_address",
            "c_comment",
        )
        .agg(
            revenue=(
                (ibis._["l_extendedprice"] * (1 - ibis._["l_discount"])).sum().round(2)
            )
        )
        .select(
            "c_custkey",
            "c_name",
            "revenue",
            "c_acctbal",
            "n_name",
            "c_address",
            "c_phone",
            "c_comment",
        )
        .order_by(ibis.desc("revenue"))
        .limit(20)
    )

    return q_final


def q11(nation, partsupp, supplier, **kwargs):
    NATION = "GERMANY"
    FRACTION = 0.0001

    q = partsupp
    q = q.join(supplier, partsupp.ps_suppkey == supplier.s_suppkey)
    q = q.join(nation, nation.n_nationkey == supplier.s_nationkey)

    q = q.filter([q.n_name == NATION])

    innerq = partsupp
    innerq = innerq.join(supplier, partsupp.ps_suppkey == supplier.s_suppkey)
    innerq = innerq.join(nation, nation.n_nationkey == supplier.s_nationkey)
    innerq = innerq.filter([innerq.n_name == NATION])
    innerq = innerq.aggregate(total=(innerq.ps_supplycost * innerq.ps_availqty).sum())

    gq = q.group_by([q.ps_partkey])
    q = gq.aggregate(value=(q.ps_supplycost * q.ps_availqty).sum())
    q = q.filter([q.value > innerq.total * FRACTION])
    q = q.order_by(ibis.desc(q.value))

    return q


def q12(orders, lineitem, **kwargs):
    var1 = "MAIL"
    var2 = "SHIP"
    var3 = date(1994, 1, 1)
    var4 = date(1995, 1, 1)

    q = orders
    q = q.join(lineitem, orders.o_orderkey == lineitem.l_orderkey)

    q = q.filter(
        [
            q.l_shipmode.isin([var1, var2]),
            q.l_commitdate < q.l_receiptdate,
            q.l_shipdate < q.l_commitdate,
            q.l_receiptdate >= ibis.date(var3),
            q.l_receiptdate < var4,
        ]
    )

    gq = q.group_by([q.l_shipmode])
    q = gq.aggregate(
        high_line_count=(
            q.o_orderpriority.case()
            .when("1-URGENT", 1)
            .when("2-HIGH", 1)
            .else_(0)
            .end()
        ).sum(),
        low_line_count=(
            q.o_orderpriority.case()
            .when("1-URGENT", 0)
            .when("2-HIGH", 0)
            .else_(1)
            .end()
        ).sum(),
    )
    q = q.order_by(q.l_shipmode)

    return q


def q13(customer, orders, **kwargs):
    var1 = "special"
    var2 = "requests"

    innerq = customer
    innerq = innerq.left_join(
        orders,
        (customer.c_custkey == orders.o_custkey)
        & ~orders.o_comment.like(f"%{var1}%{var2}%"),
    )
    innergq = innerq.group_by([innerq.c_custkey])
    innerq = innergq.aggregate(c_count=innerq.o_orderkey.count())

    gq = innerq.group_by([innerq.c_count])
    q = gq.aggregate(custdist=innerq.count())

    q = q.order_by([ibis.desc(q.custdist), ibis.desc(q.c_count)])
    return q


def q14(part, lineitem, **kwargs):
    var1 = date(1995, 9, 1)
    var2 = date(1995, 10, 1)

    q = lineitem
    q = q.join(part, lineitem.l_partkey == part.p_partkey)
    q = q.filter([q.l_shipdate >= var1, q.l_shipdate < var2])

    revenue = q.l_extendedprice * (1 - q.l_discount)
    promo_revenue = q.p_type.like("PROMO%").ifelse(revenue, 0)

    q = q.aggregate(promo_revenue=100 * promo_revenue.sum() / revenue.sum())
    return q


def q15(lineitem, supplier, **kwargs):
    var1 = date(1996, 1, 1)
    var2 = date(1996, 4, 1)

    qrev = lineitem
    qrev = qrev.filter(
        [
            lineitem.l_shipdate >= ibis.date(var1),
            lineitem.l_shipdate < var2,
        ]
    )

    gqrev = qrev.group_by([lineitem.l_suppkey])
    qrev = gqrev.aggregate(
        total_revenue=(qrev.l_extendedprice * (1 - qrev.l_discount)).sum()
    )

    q = supplier.join(qrev, supplier.s_suppkey == qrev.l_suppkey)
    q = q.filter([q.total_revenue == qrev.total_revenue.max()])
    q = q[q.s_suppkey, q.s_name, q.s_address, q.s_phone, q.total_revenue]
    return q.order_by([q.s_suppkey])


def q16(partsupp, part, supplier, **kwargs):
    BRAND = "Brand#45"
    TYPE = "MEDIUM POLISHED"
    SIZES = (49, 14, 23, 45, 19, 3, 36, 9)

    q = partsupp.join(part, part.p_partkey == partsupp.ps_partkey)
    q = q.filter(
        [
            q.p_brand != BRAND,
            ~q.p_type.like(f"{TYPE}%"),
            q.p_size.isin(SIZES),
            ~q.ps_suppkey.isin(
                supplier.filter(
                    [supplier.s_comment.like("%Customer%Complaints%")]
                ).s_suppkey
            ),
        ]
    )
    gq = q.group_by([q.p_brand, q.p_type, q.p_size])
    q = gq.aggregate(supplier_cnt=q.ps_suppkey.nunique())
    q = q.order_by([ibis.desc(q.supplier_cnt), q.p_brand, q.p_type, q.p_size])
    return q


def q17(lineitem, part, **kwargs):
    BRAND = "Brand#23"
    CONTAINER = "MED BOX"

    q = lineitem.join(part, part.p_partkey == lineitem.l_partkey)

    innerq = lineitem
    innerq = innerq.filter([innerq.l_partkey == q.p_partkey])

    q = q.filter(
        [
            q.p_brand == BRAND,
            q.p_container == CONTAINER,
            q.l_quantity < (0.2 * innerq.l_quantity.mean()),
        ]
    )
    q = q.aggregate(avg_yearly=q.l_extendedprice.sum() / 7.0)
    return q


def q18(customer, orders, lineitem, **kwargs):
    QUANTITY = 300

    subgq = lineitem.group_by([lineitem.l_orderkey])
    subq = subgq.aggregate(qty_sum=lineitem.l_quantity.sum())
    subq = subq.filter([subq.qty_sum > QUANTITY])

    q = customer
    q = q.join(orders, customer.c_custkey == orders.o_custkey)
    q = q.join(lineitem, orders.o_orderkey == lineitem.l_orderkey)
    q = q.filter([q.o_orderkey.isin(subq.l_orderkey)])

    gq = q.group_by(
        [q.c_name, q.c_custkey, q.o_orderkey, q.o_orderdate, q.o_totalprice]
    )
    q = gq.aggregate(sum_qty=q.l_quantity.sum())
    q = q.order_by([ibis.desc(q.o_totalprice), q.o_orderdate])
    return q.limit(100)


def q19(lineitem, part, **kwargs):
    QUANTITY1 = 1
    QUANTITY2 = 10
    QUANTITY3 = 20
    BRAND1 = "Brand#12"
    BRAND2 = "Brand#23"
    BRAND3 = "Brand#34"

    q = lineitem.join(part, part.p_partkey == lineitem.l_partkey)

    q1 = (
        (q.p_brand == BRAND1)
        & (q.p_container.isin(("SM CASE", "SM BOX", "SM PACK", "SM PKG")))
        & (q.l_quantity >= QUANTITY1)
        & (q.l_quantity <= QUANTITY1 + 10)
        & (q.p_size.between(1, 5))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q2 = (
        (q.p_brand == BRAND2)
        & (q.p_container.isin(("MED BAG", "MED BOX", "MED PKG", "MED PACK")))
        & (q.l_quantity >= QUANTITY2)
        & (q.l_quantity <= QUANTITY2 + 10)
        & (q.p_size.between(1, 10))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q3 = (
        (q.p_brand == BRAND3)
        & (q.p_container.isin(("LG CASE", "LG BOX", "LG PACK", "LG PKG")))
        & (q.l_quantity >= QUANTITY3)
        & (q.l_quantity <= QUANTITY3 + 10)
        & (q.p_size.between(1, 15))
        & (q.l_shipmode.isin(("AIR", "AIR REG")))
        & (q.l_shipinstruct == "DELIVER IN PERSON")
    )

    q = q.filter([q1 | q2 | q3])
    q = q.aggregate(revenue=(q.l_extendedprice * (1 - q.l_discount)).sum())
    return q


def q20(supplier, nation, partsupp, part, lineitem, **kwargs):
    var1 = date(1994, 1, 1)
    var2 = date(1995, 1, 1)
    var3 = "CANADA"
    var4 = "forest"

    q1 = supplier.join(nation, supplier.s_nationkey == nation.n_nationkey)

    q3 = part.filter([part.p_name.like(f"{var4}%")])
    q2 = partsupp

    q4 = lineitem.filter(
        [
            lineitem.l_partkey == q2.ps_partkey,
            lineitem.l_suppkey == q2.ps_suppkey,
            lineitem.l_shipdate >= var1,
            lineitem.l_shipdate < var2,
        ]
    )

    q2 = q2.filter(
        [
            partsupp.ps_partkey.isin(q3.p_partkey),
            partsupp.ps_availqty > 0.5 * q4.l_quantity.sum(),
        ]
    )

    q1 = q1.filter([q1.n_name == var3, q1.s_suppkey.isin(q2.ps_suppkey)])

    q1 = q1[q1.s_name, q1.s_address]

    return q1.order_by(q1.s_name)


def q21(supplier, lineitem, orders, nation, **kwargs):
    NATION = "SAUDI ARABIA"

    L2 = lineitem.view()
    L3 = lineitem.view()

    q = supplier
    q = q.join(lineitem, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(orders, orders.o_orderkey == lineitem.l_orderkey)
    q = q.join(nation, supplier.s_nationkey == nation.n_nationkey)
    q = q[
        q.l_orderkey.name("l1_orderkey"),
        q.o_orderstatus,
        q.l_receiptdate,
        q.l_commitdate,
        q.l_suppkey.name("l1_suppkey"),
        q.s_name,
        q.n_name,
    ]
    q = q.filter(
        [
            q.o_orderstatus == "F",
            q.l_receiptdate > q.l_commitdate,
            q.n_name == NATION,
            ((L2.l_orderkey == q.l1_orderkey) & (L2.l_suppkey != q.l1_suppkey)).any(),
            ~(
                (
                    (L3.l_orderkey == q.l1_orderkey)
                    & (L3.l_suppkey != q.l1_suppkey)
                    & (L3.l_receiptdate > L3.l_commitdate)
                ).any()
            ),
        ]
    )

    gq = q.group_by([q.s_name])
    q = gq.aggregate(numwait=q.count())
    q = q.order_by([ibis.desc(q.numwait), q.s_name])
    return q.limit(100)


def q22(customer, orders, **kwargs):
    COUNTRY_CODES = ("13", "31", "23", "29", "30", "18", "17")

    q = customer.filter(
        [
            customer.c_acctbal > 0.00,
            customer.c_phone.substr(0, 2).isin(COUNTRY_CODES),
        ]
    )
    q = q.aggregate(avg_bal=customer.c_acctbal.mean())

    custsale = customer.filter(
        [
            customer.c_phone.substr(0, 2).isin(COUNTRY_CODES),
            customer.c_acctbal > q.avg_bal,
            ~(orders.o_custkey == customer.c_custkey).any(),
        ]
    )
    custsale = custsale[
        customer.c_phone.substr(0, 2).name("cntrycode"), customer.c_acctbal
    ]

    gq = custsale.group_by(custsale.cntrycode)
    outerq = gq.aggregate(numcust=custsale.count(), totacctbal=custsale.c_acctbal.sum())

    return outerq.order_by(outerq.cntrycode)


all_queries = [
    q1,
    q2,
    q3,
    q4,
    q5,
    q6,
    q7,
    q8,
    q9,
    q10,
    q11,
    q12,
    q13,
    q14,
    q15,
    q16,
    q17,
    q18,
    q19,
    q20,
    q21,
    q22,
]



import ibis
import gcsfs
import tomllib
import streamlit as st
import plotly.express as px

from ibis_bench.utils.monitor import get_cache_dir

st.set_page_config(layout="wide")
st.title("Ibis benchmarking results")

with open("pyproject.toml", "rb") as f:
    pyproject = tomllib.load(f)

dependencies = pyproject["project"]["dependencies"]

datafusion_version = [d for d in dependencies if "datafusion" in d][0]
duckdb_version = [d for d in dependencies if "duckdb" in d][0]
polars_version = [d for d in dependencies if "polars" in d][0]
ibis_version = [d for d in dependencies if "ibis" in d][0]

details = f"""
[blog blog blog](https://ibis-project.org/posts/ibis-bench)

the purpose of this dashboard is to compare TPC-H benchmarks across the big three single-node, Apache Arrow-based, modern OLAP engines: DuckDB, DataFusion, and Polars

versions:

- DataFusion: `{datafusion_version}`
- DuckDB: `{duckdb_version}`
- Polars: `{polars_version}`
- Ibis: `{ibis_version}`
"""
details = details.strip()
st.markdown(details)

with open("results.py") as f:
    code = f.read()

with st.expander("Show source code for this page", expanded=False):
    st.code(code, line_numbers=True, language="python")

ibis.options.interactive = True
ibis.options.repr.interactive.max_rows = 20
ibis.options.repr.interactive.max_columns = None

# dark mode for px
px.defaults.template = "plotly_dark"


def get_t():
    # ibis connection
    # con = ibis.connect("duckdb://cache.ddb")
    con = ibis.connect("duckdb://")

    # cloud logs
    cloud = False

    if cloud:
        PROJECT = "voltrondata-demo"
        BUCKET = "ibis-bench"
        # BUCKET = "ibis-benchy"

        fs = gcsfs.GCSFileSystem(project=PROJECT)

        con.register_filesystem(fs)

        glob = f"gs://{BUCKET}/{get_cache_dir()}/*.parquet"
    else:
        glob = f"bench_logs_v2/raw_json/*.json"

    print(glob)

    # read data
    table_name = "bench_data"
    if table_name not in con.list_tables():
        t = (
            con.read_json(glob)
            .mutate(
                timestamp=ibis._["timestamp"].cast("timestamp"),
            )
            .drop("file_id")
            .distinct()  # TODO: hmmmmmm
            .filter(ibis._["file_type"] == "parquet")  # TODO: remove after CSV runs
            .mutate(instance_type=ibis.literal("work laptop"))
            .cache()
        )
        con.create_table(table_name, t)
    else:
        t = con.table(table_name)

    return t


def get_instance_details(t):
    cpu_type_cases = (
        ibis.case()
        .when(
            ibis._["instance_type"].startswith("n2d"),
            "AMD EPYC",
        )
        .when(
            ibis._["instance_type"].startswith("n2"),
            "Intel Cascade and Ice Lake",
        )
        .when(
            ibis._["instance_type"].startswith("c3"),
            "Intel Saphire Rapids",
        )
        .when(
            ibis._["instance_type"] == "work laptop",
            "Apple M1 Max",
        )
        .when(
            ibis._["instance_type"] == "personal laptop",
            "Apple M2 Max",
        )
        .else_("unknown")
        .end()
    )
    cpu_num_cases = (
        ibis.case()
        .when(
            ibis._["instance_type"].contains("-"),
            ibis._["instance_type"].split("-")[-1].cast("int"),
        )
        .when(ibis._["instance_type"].contains("laptop"), 12)
        .else_(0)
        .end()
    )
    memory_gb_cases = (
        ibis.case()
        .when(
            ibis._["instance_type"].contains("-"),
            ibis._["instance_type"].split("-")[-1].cast("int") * 4,
        )
        .when(ibis._["instance_type"] == "work laptop", 32)
        .when(ibis._["instance_type"] == "personal laptop", 96)
        .else_(0)
        .end()
    )

    instance_details = (
        t.group_by("instance_type")
        .agg()
        .mutate(
            cpu_type=cpu_type_cases, cpu_cores=cpu_num_cases, memory_gbs=memory_gb_cases
        )
    ).order_by("memory_gbs", "cpu_cores", "instance_type")

    return instance_details


t = get_t()
instance_details = get_instance_details(t)


# streamlit viz beyond this point
def totals_metrics(t):
    cols = st.columns(6)
    with cols[0]:
        st.metric(
            label="total queries run",
            value=f"{t.count().to_pandas():,}",
        )
    with cols[1]:
        st.metric(
            label="total runtime minutes",
            value=f"{round(t['execution_seconds'].sum().to_pandas() / 60, 2):,}",
            help=f"average: {round(t['execution_seconds'].mean().to_pandas(), 2)}s/query",
        )
    with cols[2]:
        st.metric(
            label="total systems",
            value=t.select("system").distinct().count().to_pandas(),
        )
    with cols[3]:
        st.metric(
            label="total instance types",
            value=t.select("instance_type").distinct().count().to_pandas(),
        )
    with cols[4]:
        st.metric(
            label="total scale factors",
            value=t.select("sf").distinct().count().to_pandas(),
        )
    with cols[5]:
        st.metric(
            label="total queries",
            value=t.select("query_number").distinct().count().to_pandas(),
        )


st.markdown("## totals (all data)")
totals_metrics(t)

# user options
st.markdown("## data filters")
with st.form(key="app"):
    # system options
    system_options = sorted(
        t.select("system").distinct().to_pandas()["system"].tolist()
    )
    systems = st.multiselect(
        "select system(s)",
        system_options,
        # default=system_options,
        default=["ibis-datafusion", "ibis-duckdb", "polars-lazy"],
    )

    instance_type_options = sorted(
        t.select("instance_type").distinct().to_pandas()["instance_type"].tolist(),
        key=lambda x: (x.split("-")[0], int(x.split("-")[-1]))
        if "-" in x
        else ("z" + x[3], 0),
    )
    instance_types = st.multiselect(
        "select instance type(s)",
        instance_type_options,
        # default=[
        #     instance
        #     for instance in instance_type_options
        #     if instance.startswith("n2d")
        #     # instance
        #     # for instance in instance_type_options
        #     # if "laptop" in instance
        # ],
        default=instance_type_options,
    )
    instance_types = sorted(
        instance_types,
        key=lambda x: (x.split("-")[0], int(x.split("-")[-1]))
        if "-" in x
        else ("z" + x[3], 0),
    )

    sfs = sorted(t.select("sf").distinct().to_pandas()["sf"].tolist())
    scale_factor = st.radio(
        "select scale factor",
        sfs,
        index=sfs.index(sfs[-1]),
    )

    # filetype options
    # filetype_options = sorted(
    #     t.select("file_type").distinct().to_pandas()["file_type"].tolist()
    # )
    # file_type = st.radio(
    #     "select file type",
    #     filetype_options,
    #     index=filetype_options.index("parquet") if "parquet" in filetype_options else 0,
    # )
    file_type = "parquet"

    # query options
    query_numbers = sorted(
        t.select("query_number").distinct().to_pandas()["query_number"].tolist()
    )
    start_query, end_query = st.select_slider(
        "select a range of queries",
        options=query_numbers,
        value=(min(query_numbers), max(query_numbers)),
    )
    query_numbers = list(range(start_query, end_query + 1))

    # log_y
    log_y = st.toggle("log y-axis", True)

    # submit button
    update_button = st.form_submit_button(label="update")

st.markdown("## totals (filtered data)")
totals_metrics(
    t.filter(t["sf"] == scale_factor)
    .filter(t["system"].isin(systems))
    .filter(t["file_type"] == file_type)
    .filter(t["instance_type"].isin(instance_types))
    .filter(t["query_number"] >= start_query)
    .filter(t["query_number"] <= end_query)
)

# aggregate data
agg = (
    t.filter(t["sf"] == scale_factor)
    .filter(t["system"].isin(systems))
    # .filter(t["file_type"].isin(file_type))
    .filter(t["file_type"] == file_type)
    .filter(t["instance_type"].isin(instance_types))
    .filter(t["query_number"] >= start_query)
    .filter(t["query_number"] <= end_query)
    .group_by("system", "instance_type", "sf", "query_number")  # , "file_type")
    .agg(
        mean_execution_seconds=t["execution_seconds"].mean(),
    )
    .order_by(
        ibis.desc("sf"),
        ibis.asc("query_number"),
        ibis.desc("system"),
        ibis.desc("instance_type"),
        ibis.asc("mean_execution_seconds"),
        # ibis.asc("file_type"),
    )
)
agg = agg.join(instance_details, "instance_type")

all_systems = sorted(t.select("system").distinct().to_pandas()["system"].tolist())

category_orders = {
    "query_number": sorted(query_numbers),
    "system": sorted(all_systems),
    "instance_type": instance_types,
}

sfs = agg.select("sf").distinct().to_pandas()["sf"].tolist()
gb_factor = 2 / 5 if file_type == "parquet" else 11 / 10

for sf in sorted(sfs):
    st.markdown("## execution time per query")
    c = px.bar(
        agg,
        x="query_number",
        y="mean_execution_seconds",
        log_y=log_y,
        color="system",
        category_orders=category_orders,
        barmode="group",
        pattern_shape="instance_type",
        hover_data=["cpu_type", "cpu_cores", "memory_gbs"],
        title=f"scale factor: {sf} (~{sf}GB in memory | ~{round(sf * gb_factor, 2)}GB as {file_type})",
    )
    st.plotly_chart(c)

    agg2 = agg.group_by("system", "instance_type").agg(
        present_queries=ibis._["query_number"].collect().unique().sort(),
        total_mean_execution_seconds=ibis._["mean_execution_seconds"].sum(),
    )
    agg2 = agg2.join(instance_details, "instance_type")
    agg2 = (
        agg2.mutate(
            failing_queries=t.filter(t["query_number"].isin(query_numbers))
            .distinct(on="query_number")["query_number"]
            .collect()
            .filter(lambda x: ~agg2["present_queries"].contains(x))
        )
        .mutate(
            num_failing_queries=ibis._["failing_queries"].length(),
            num_successful_queries=ibis._["present_queries"].length(),
        )
        .drop("present_queries")
        .order_by(ibis.desc("memory_gbs"), "system")
    )

    st.markdown("## completed queries by system and instance type")
    c = px.bar(
        agg2,
        x="system",
        y="num_successful_queries",
        color="instance_type",
        barmode="group",
        hover_data=["cpu_type", "cpu_cores", "memory_gbs"],
        category_orders={
            "system": sorted(systems),
            "instance_type": reversed(instance_types),
        },
        title="completed queries",
    )
    st.plotly_chart(c)

    st.markdown("## details by system and instance type")
    st.dataframe(agg2, use_container_width=True)

    st.markdown("---")



import os
import subprocess

instance_type = "work laptop"
all_systems = [
    "ibis-duckdb",
    "ibis-duckdb-sql",
    "ibis-datafusion",
    "ibis-datafusion-sql",
    "polars-lazy",
    "ibis-polars",
]
all_sfs = [
    1,
    8,
    16,
    32,
    64,
    128,
]
all_qs = range(1, 23)


def main():
    for system in all_systems:
        for sf in all_sfs:
            for q in all_qs:
                cmd = f"bench run {system} -s {sf} -q {q} -i '{instance_type}'"

                print(f"running: {cmd}")
                res = subprocess.run(cmd, shell=True)
                if res.returncode != 0:
                    print(f"failed to run: {cmd}")

                print(f"finished running: {cmd}")


if __name__ == "__main__":
    main()



import os
import ibis
import polars as pl

from ibis_bench.utils.logging import log


def write_results(
    res,
    sf: int,
    n_partitions: int,
    system: str,
    q_number: int,
):
    dirname = os.path.join("results_data", system)

    if not os.path.exists(dirname):
        os.makedirs(dirname, exist_ok=True)

    if isinstance(res, ibis.Table):
        log.info(
            f"\twriting ibis result to {os.path.join(dirname, f'q_{q_number}.parquet')}..."
        )
        res.to_parquet(os.path.join(dirname, f"q_{q_number}.parquet"))
        log.info(
            f"\tdone writing ibis result to {os.path.join(dirname, f'q_{q_number}.parquet')}..."
        )
    elif isinstance(res, pl.DataFrame):
        log.info(
            f"\twriting polars (eager) result to {os.path.join(dirname, f'q_{q_number}.parquet')}..."
        )
        res.write_parquet(os.path.join(dirname, f"q_{q_number}.parquet"))
        log.info(
            f"\tdone writing polars (eager) result to {os.path.join(dirname, f'q_{q_number}.parquet')}..."
        )
    elif isinstance(res, pl.LazyFrame):
        log.info(
            f"\twriting polars (lazy) result to {os.path.join(dirname, f'q_{q_number}.parquet')}..."
        )
        res.collect().write_parquet(os.path.join(dirname, f"q_{q_number}.parquet"))
        log.info(
            f"\tdone writing polars (lazy) result to {os.path.join(dirname, f'q_{q_number}.parquet')}..."
        )

        # NOTE: sinking parquet caused multiple issues, including: https://github.com/pola-rs/polars/issues/16694
        # for now, we'll just use collect + write_parquet and revisit later when the streaming engine is redone
        # try:
        #     # https://github.com/pola-rs/polars/issues/6603
        #     # InvalidOperationError: sink_Parquet(ParquetWriteOptions { compression: Zstd(None), statistics: true, row_group_size: None, data_pagesize_limit: None, maintain_order: true }) not yet supported in standard engine. Use 'collect().write_parquet()'
        #     # "not all queries are supported for sinking"
        #     # they should fail instantly, so we can just use collect() + write_parquet if this fails
        #     res.sink_parquet(os.path.join(dirname, f"q_{q_number}.parquet"))
        # except:

    del res



import os
import sys
import logging as log

from datetime import datetime

# TODO: this is all very hacky
file_logging = False

if file_logging:
    log_name = str(datetime.now().strftime("%Y-%m-%d:%s"))
    dir_name = "bench_cli_logs"

    if not os.path.exists(dir_name):
        os.makedirs(dir_name, exist_ok=True)

    log.basicConfig(
        level=log.INFO,
        filename=os.path.join(dir_name, f"ibis_bench_{log_name}.log"),
        filemode="a",
        format=" %(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    # sys.stdout = open(
    #    os.path.join(dir_name, f"ibis_bench_{log_name}_stdout.log"),
    #    "a",
    # )
    sys.stderr = open(
        os.path.join(dir_name, f"ibis_bench_{log_name}_stderr.log"),
        "a",
    )

else:
    log.basicConfig(level=log.INFO)



import os
import ibis
import uuid
import typer
import subprocess

from ibis_bench.utils.logging import log
from ibis_bench.utils.monitor import get_raw_json_dir, get_cache_dir

DEFAULT_INSTANCE_TYPE = "unknown"
DEFAULT_SYSTEMS = [
    "ibis-datafusion",
    "ibis-datafusion-sql",
    "ibis-duckdb",
    "ibis-duckdb-sql",
    "ibis-polars",
    "polars-lazy",
]
DEFAULT_SFS = [
    1,
    8,
    16,
    32,
    64,
    128,
]
DEFAULT_QS = range(1, 23)

TYPER_KWARGS = {
    "no_args_is_help": True,
    "add_completion": False,
    "context_settings": {"help_option_names": ["-h", "--help"]},
}

app = typer.Typer(help="ibis-bench", **TYPER_KWARGS)


@app.command()
def run(
    systems: list[str] = typer.Option(
        DEFAULT_SYSTEMS, "--system", "-S", help="system to run on"
    ),
    scale_factors: list[int] = typer.Option(
        DEFAULT_SFS, "--scale-factor", "-s", help="scale factors"
    ),
    queries: list[int] = typer.Option(
        DEFAULT_QS, "--query-number", "-q", help="query numbers"
    ),
    instance_type: str = typer.Option(
        DEFAULT_INSTANCE_TYPE, "--instance-type", "-i", help="instance type"
    ),
    use_csv: bool = typer.Option(
        False, "--csv", "-c", help="use CSV instead of Parquet"
    ),
    repeat: int = typer.Option(
        3, "--repeat", "-r", help="number of times to repeat the run"
    ),
):
    """
    run bench
    """
    for _ in range(repeat):
        for sf in scale_factors:
            for system in systems:
                for q in queries:
                    cmd = f"bench run {system} -s {sf} -q {q} -i '{instance_type}'"
                    cmd += " --csv" if use_csv else ""

                    log.info(f"running: {cmd}")
                    res = subprocess.run(cmd, shell=True)
                    if res.returncode != 0:
                        log.info(f"failed to run: {cmd}")

                    log.info(f"finished running: {cmd}")


@app.command()
def combine_json(instance_type: str = typer.Argument(..., help="instance type")):
    """
    combine JSON files as Parquet
    """
    file_id = str(uuid.uuid4())

    con = ibis.connect("duckdb://")
    t = con.read_json(f"{get_raw_json_dir()}/file_id=*.json")
    t = t.mutate(instance_type=ibis.literal(instance_type))

    file_path = os.path.join(get_cache_dir(), f"file_id={file_id}.parquet")

    log.info(f"\twriting monitor data to {file_path}...")
    t.to_parquet(file_path)
    log.info(f"\tdone writing monitor data to {file_path}...")


if __name__ == "__main__":
    typer.run(app)



import os
import ibis

from ibis_bench.utils.logging import log


def get_data_dir(sf, csv: bool = False):
    dir_name = "csv" if csv else "parquet"
    return os.path.join("tpcds_data", dir_name, f"sf={sf}")


def generate_data(sf, csv: bool = False):
    log.info(f"generating data for sf={sf}...")

    con = ibis.connect("duckdb://")
    con.raw_sql("PRAGMA disable_progress_bar;")

    parquet_data_directory = get_data_dir(sf)
    csv_data_directory = get_data_dir(sf, csv=True)

    if not os.path.exists(parquet_data_directory):
        con.raw_sql(f"call dsdgen(sf={sf})")
        for table in con.list_tables():
            os.makedirs(os.path.join(parquet_data_directory, table), exist_ok=True)
            if csv:
                os.makedirs(os.path.join(csv_data_directory, table), exist_ok=True)
            log.info(
                f"\twriting {os.path.join(parquet_data_directory, table, f'{0:04d}.parquet')}..."
            )
            con.table(table).to_parquet(
                os.path.join(parquet_data_directory, table, f"{0:04d}.parquet")
            )
            log.info(
                f"\tdone writing {os.path.join(parquet_data_directory, table, f'{0:04d}.parquet')}"
            )
            if csv:
                log.info(
                    f"\twriting {os.path.join(csv_data_directory, table, f'{0:04d}.csv')}..."
                )
                con.table(table).to_csv(
                    os.path.join(csv_data_directory, table, f"{0:04d}.csv")
                )
                log.info(
                    f"\tdone writing {os.path.join(csv_data_directory, table, f'{0:04d}.csv')}"
                )

            con.drop_table(table)
    else:
        log.info(f"\tdata already exists at {parquet_data_directory}, skipping...")


def _parquets_to_csvs():
    con = ibis.connect("duckdb://")
    for root, _, files in os.walk(os.path.join("tpcds_data", "parquet")):
        for file in files:
            if file.endswith(".parquet"):
                parquet_file = os.path.join(root, file)
                csv_file = parquet_file.replace("parquet", "csv")

                os.makedirs(os.path.dirname(csv_file), exist_ok=True)

                log.info(f"reading {parquet_file}...")
                t = con.read_parquet(parquet_file)

                log.info(f"writing {csv_file}...")
                t.to_csv(csv_file)



import os
import ibis
import json
import time
import uuid

from datetime import datetime

from ibis_bench.utils.logging import log
from ibis_bench.utils.write_data import write_results


def get_timings_dir():
    dir_name = "bench_logs_v2"

    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    return dir_name


def get_raw_json_dir():
    dir_name = os.path.join(get_timings_dir(), "raw_json")

    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    return dir_name


def get_cache_dir():
    dir_name = os.path.join(get_timings_dir(), "cache")

    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    return dir_name


def monitor_it(
    func,
    sf: int,
    n_partitions: int,
    query_number: int,
    system: str,
    session_id: str,
    instance_type: str,
    use_csv: bool,
    decimal_to_float: bool,
    *args,
    **kwargs,
):
    log.info(
        f"running and monitoring for system {system} query {query_number} at scale factor {sf} and {n_partitions} partitions (session id {session_id})..."
    )
    start_time = time.time()
    write_results(func(*args, **kwargs), sf, n_partitions, system, query_number)
    elapsed_time = time.time() - start_time
    log.info(
        f"done running and monitoring for system {system} query {query_number} at scale factor {sf} and {n_partitions} partitions (session id {session_id})..."
    )

    data = {
        "session_id": session_id,
        "instance_type": instance_type,
        "system": system,
        "timestamp": datetime.utcnow().isoformat(),
        "sf": sf,
        "n_partitions": n_partitions,
        "query_number": query_number,
        "execution_seconds": elapsed_time,
        "file_type": "csv" if use_csv else "parquet",
        "floats": decimal_to_float,
    }

    write_monitor_results(data)


def write_monitor_results(results):
    file_id = str(uuid.uuid4())
    file_path = os.path.join(get_raw_json_dir(), f"file_id={file_id}.json")

    log.info(f"\twriting monitor data to {file_path}...")
    with open(file_path, "w") as f:
        json.dump(results, f)
    log.info(f"\tdone writing monitor data to {file_path}...")



# adapted from: https://github.com/ibis-project/ibis/tree/main/ibis/backends/tests/tpch

q1_sql = """
SELECT
    l_returnflag,
    l_linestatus,
    sum(l_quantity) AS sum_qty,
    sum(l_extendedprice) AS sum_base_price,
    sum(l_extendedprice * (1 - l_discount)) AS sum_disc_price,
    sum(l_extendedprice * (1 - l_discount) * (1 + l_tax)) AS sum_charge,
    avg(l_quantity) AS avg_qty,
    avg(l_extendedprice) AS avg_price,
    avg(l_discount) AS avg_disc,
    count(*) AS count_order
FROM
    lineitem
WHERE
    l_shipdate <= CAST('1998-09-02' AS date)
GROUP BY
    l_returnflag,
    l_linestatus
ORDER BY
    l_returnflag,
    l_linestatus;
"""
q1_sql = q1_sql.strip().strip(";")


def q1(lineitem, dialect="duckdb", **kwargs):
    return lineitem.sql(q1_sql, dialect=dialect)


q2_sql = """
SELECT
    s_acctbal,
    s_name,
    n_name,
    p_partkey,
    p_mfgr,
    s_address,
    s_phone,
    s_comment
FROM
    part,
    supplier,
    partsupp,
    nation,
    region
WHERE
    p_partkey = ps_partkey
    AND s_suppkey = ps_suppkey
    AND p_size = 15
    AND p_type LIKE '%BRASS'
    AND s_nationkey = n_nationkey
    AND n_regionkey = r_regionkey
    AND r_name = 'EUROPE'
    AND ps_supplycost = (
        SELECT
            min(ps_supplycost)
        FROM
            partsupp,
            supplier,
            nation,
            region
        WHERE
            p_partkey = ps_partkey
            AND s_suppkey = ps_suppkey
            AND s_nationkey = n_nationkey
            AND n_regionkey = r_regionkey
            AND r_name = 'EUROPE')
ORDER BY
    s_acctbal DESC,
    n_name,
    s_name,
    p_partkey
LIMIT 100;
"""
q2_sql = q2_sql.strip().strip(";")


def q2(nation, part, partsupp, region, supplier, dialect="duckdb", **kwargs):
    return part.sql(q2_sql, dialect=dialect)


q3_sql = """
SELECT
    l_orderkey,
    sum(l_extendedprice * (1 - l_discount)) AS revenue,
    o_orderdate,
    o_shippriority
FROM
    customer,
    orders,
    lineitem
WHERE
    c_mktsegment = 'BUILDING'
    AND c_custkey = o_custkey
    AND l_orderkey = o_orderkey
    AND o_orderdate < CAST('1995-03-15' AS date)
    AND l_shipdate > CAST('1995-03-15' AS date)
GROUP BY
    l_orderkey,
    o_orderdate,
    o_shippriority
ORDER BY
    revenue DESC,
    o_orderdate
LIMIT 10;
"""
q3_sql = q3_sql.strip().strip(";")


def q3(customer, lineitem, orders, dialect="duckdb", **kwargs):
    return customer.sql(q3_sql, dialect=dialect)


q4_sql = """
SELECT
    o_orderpriority,
    count(*) AS order_count
FROM
    orders
WHERE
    o_orderdate >= CAST('1993-07-01' AS date)
    AND o_orderdate < CAST('1993-10-01' AS date)
    AND EXISTS (
        SELECT
            *
        FROM
            lineitem
        WHERE
            l_orderkey = o_orderkey
            AND l_commitdate < l_receiptdate)
GROUP BY
    o_orderpriority
ORDER BY
    o_orderpriority;
"""
q4_sql = q4_sql.strip().strip(";")


def q4(lineitem, orders, dialect="duckdb", **kwargs):
    return orders.sql(q4_sql, dialect=dialect)


q5_sql = """
SELECT
    n_name,
    sum(l_extendedprice * (1 - l_discount)) AS revenue
FROM
    customer,
    orders,
    lineitem,
    supplier,
    nation,
    region
WHERE
    c_custkey = o_custkey
    AND l_orderkey = o_orderkey
    AND l_suppkey = s_suppkey
    AND c_nationkey = s_nationkey
    AND s_nationkey = n_nationkey
    AND n_regionkey = r_regionkey
    AND r_name = 'ASIA'
    AND o_orderdate >= CAST('1994-01-01' AS date)
    AND o_orderdate < CAST('1995-01-01' AS date)
GROUP BY
    n_name
ORDER BY
    revenue DESC;
"""
q5_sql = q5_sql.strip().strip(";")


def q5(
    region, nation, supplier, customer, orders, lineitem, dialect="duckdb", **kwargs
):
    return region.sql(q5_sql, dialect=dialect)


q6_sql = """
SELECT
    sum(l_extendedprice * l_discount) AS revenue
FROM
    lineitem
WHERE
    l_shipdate >= CAST('1994-01-01' AS date)
    AND l_shipdate < CAST('1995-01-01' AS date)
    AND l_discount BETWEEN 0.05
    AND 0.07
    AND l_quantity < 24;
"""
q6_sql = q6_sql.strip().strip(";")


def q6(lineitem, dialect="duckdb", **kwargs):
    return lineitem.sql(q6_sql, dialect=dialect)


q7_sql = """
SELECT
    supp_nation,
    cust_nation,
    l_year,
    sum(volume) AS revenue
FROM (
    SELECT
        n1.n_name AS supp_nation,
        n2.n_name AS cust_nation,
        extract(year FROM l_shipdate) AS l_year,
        l_extendedprice * (1 - l_discount) AS volume
    FROM
        supplier,
        lineitem,
        orders,
        customer,
        nation n1,
        nation n2
    WHERE
        s_suppkey = l_suppkey
        AND o_orderkey = l_orderkey
        AND c_custkey = o_custkey
        AND s_nationkey = n1.n_nationkey
        AND c_nationkey = n2.n_nationkey
        AND ((n1.n_name = 'FRANCE'
                AND n2.n_name = 'GERMANY')
            OR (n1.n_name = 'GERMANY'
                AND n2.n_name = 'FRANCE'))
        AND l_shipdate BETWEEN CAST('1995-01-01' AS date)
        AND CAST('1996-12-31' AS date)) AS shipping
GROUP BY
    supp_nation,
    cust_nation,
    l_year
ORDER BY
    supp_nation,
    cust_nation,
    l_year;
"""
q7_sql = q7_sql.strip().strip(";")


def q7(nation, customer, orders, lineitem, supplier, dialect="duckdb", **kwargs):
    return nation.sql(q7_sql, dialect=dialect)


q8_sql = """
SELECT
    o_year,
    sum(
        CASE WHEN nation = 'BRAZIL' THEN
            volume
        ELSE
            0
        END) / sum(volume) AS mkt_share
FROM (
    SELECT
        extract(year FROM o_orderdate) AS o_year,
        l_extendedprice * (1 - l_discount) AS volume,
        n2.n_name AS nation
    FROM
        part,
        supplier,
        lineitem,
        orders,
        customer,
        nation n1,
        nation n2,
        region
    WHERE
        p_partkey = l_partkey
        AND s_suppkey = l_suppkey
        AND l_orderkey = o_orderkey
        AND o_custkey = c_custkey
        AND c_nationkey = n1.n_nationkey
        AND n1.n_regionkey = r_regionkey
        AND r_name = 'AMERICA'
        AND s_nationkey = n2.n_nationkey
        AND o_orderdate BETWEEN CAST('1995-01-01' AS date)
        AND CAST('1996-12-31' AS date)
        AND p_type = 'ECONOMY ANODIZED STEEL') AS all_nations
GROUP BY
    o_year
ORDER BY
    o_year;
"""
q8_sql = q8_sql.strip().strip(";")


def q8(
    nation,
    customer,
    orders,
    lineitem,
    supplier,
    part,
    region,
    dialect="duckdb",
    **kwargs,
):
    return part.sql(q8_sql, dialect=dialect)


q9_sql = """
SELECT
    nation,
    o_year,
    sum(amount) AS sum_profit
FROM (
    SELECT
        n_name AS nation,
        extract(year FROM o_orderdate) AS o_year,
        l_extendedprice * (1 - l_discount) - ps_supplycost * l_quantity AS amount
    FROM
        part,
        supplier,
        lineitem,
        partsupp,
        orders,
        nation
    WHERE
        s_suppkey = l_suppkey
        AND ps_suppkey = l_suppkey
        AND ps_partkey = l_partkey
        AND p_partkey = l_partkey
        AND o_orderkey = l_orderkey
        AND s_nationkey = n_nationkey
        AND p_name LIKE '%green%') AS profit
GROUP BY
    nation,
    o_year
ORDER BY
    nation,
    o_year DESC;
"""
q9_sql = q9_sql.strip().strip(";")


def q9(lineitem, nation, orders, part, partsupp, supplier, dialect="duckdb", **kwargs):
    return lineitem.sql(q9_sql, dialect=dialect)


q10_sql = """
SELECT
    c_custkey,
    c_name,
    sum(l_extendedprice * (1 - l_discount)) AS revenue,
    c_acctbal,
    n_name,
    c_address,
    c_phone,
    c_comment
FROM
    customer,
    orders,
    lineitem,
    nation
WHERE
    c_custkey = o_custkey
    AND l_orderkey = o_orderkey
    AND o_orderdate >= CAST('1993-10-01' AS date)
    AND o_orderdate < CAST('1994-01-01' AS date)
    AND l_returnflag = 'R'
    AND c_nationkey = n_nationkey
GROUP BY
    c_custkey,
    c_name,
    c_acctbal,
    c_phone,
    n_name,
    c_address,
    c_comment
ORDER BY
    revenue DESC
LIMIT 20;
"""
q10_sql = q10_sql.strip().strip(";")


def q10(customer, lineitem, nation, orders, dialect="duckdb", **kwargs):
    return customer.sql(q10_sql, dialect=dialect)


q11_sql = """
SELECT
    ps_partkey,
    sum(ps_supplycost * ps_availqty) AS value
FROM
    partsupp,
    supplier,
    nation
WHERE
    ps_suppkey = s_suppkey
    AND s_nationkey = n_nationkey
    AND n_name = 'GERMANY'
GROUP BY
    ps_partkey
HAVING
    sum(ps_supplycost * ps_availqty) > (
        SELECT
            sum(ps_supplycost * ps_availqty) * 0.0001000000
        FROM
            partsupp,
            supplier,
            nation
        WHERE
            ps_suppkey = s_suppkey
            AND s_nationkey = n_nationkey
            AND n_name = 'GERMANY')
ORDER BY
    value DESC;
"""
q11_sql = q11_sql.strip().strip(";")


def q11(nation, partsupp, supplier, dialect="duckdb", **kwargs):
    return partsupp.sql(q11_sql, dialect=dialect)


q12_sql = """
SELECT
    l_shipmode,
    sum(
        CASE WHEN o_orderpriority = '1-URGENT'
            OR o_orderpriority = '2-HIGH' THEN
            1
        ELSE
            0
        END) AS high_line_count,
    sum(
        CASE WHEN o_orderpriority <> '1-URGENT'
            AND o_orderpriority <> '2-HIGH' THEN
            1
        ELSE
            0
        END) AS low_line_count
FROM
    orders,
    lineitem
WHERE
    o_orderkey = l_orderkey
    AND l_shipmode IN ('MAIL', 'SHIP')
    AND l_commitdate < l_receiptdate
    AND l_shipdate < l_commitdate
    AND l_receiptdate >= CAST('1994-01-01' AS date)
    AND l_receiptdate < CAST('1995-01-01' AS date)
GROUP BY
    l_shipmode
ORDER BY
    l_shipmode;
"""
q12_sql = q12_sql.strip().strip(";")


def q12(orders, lineitem, dialect="duckdb", **kwargs):
    return orders.sql(q12_sql, dialect=dialect)


q13_sql = """
SELECT
    c_count,
    count(*) AS custdist
FROM (
    SELECT
        c_custkey,
        count(o_orderkey)
    FROM
        customer
    LEFT OUTER JOIN orders ON c_custkey = o_custkey
    AND o_comment NOT LIKE '%special%requests%'
GROUP BY
    c_custkey) AS c_orders (c_custkey,
        c_count)
GROUP BY
    c_count
ORDER BY
    custdist DESC,
    c_count DESC;
"""
q13_sql = q13_sql.strip().strip(";")


def q13(customer, orders, dialect="duckdb", **kwargs):
    return customer.sql(q13_sql, dialect=dialect)


q14_sql = """
SELECT
    100.00 * sum(
        CASE WHEN p_type LIKE 'PROMO%' THEN
            l_extendedprice * (1 - l_discount)
        ELSE
            0
        END) / sum(l_extendedprice * (1 - l_discount)) AS promo_revenue
FROM
    lineitem,
    part
WHERE
    l_partkey = p_partkey
    AND l_shipdate >= date '1995-09-01'
    AND l_shipdate < CAST('1995-10-01' AS date);
"""
q14_sql = q14_sql.strip().strip(";")


def q14(lineitem, part, dialect="duckdb", **kwargs):
    return lineitem.sql(q14_sql, dialect=dialect)


q15_sql = """
SELECT
    s_suppkey,
    s_name,
    s_address,
    s_phone,
    total_revenue
FROM
    supplier,
    (
        SELECT
            l_suppkey AS supplier_no,
            sum(l_extendedprice * (1 - l_discount)) AS total_revenue
        FROM
            lineitem
        WHERE
            l_shipdate >= CAST('1996-01-01' AS date)
            AND l_shipdate < CAST('1996-04-01' AS date)
        GROUP BY
            supplier_no) revenue0
WHERE
    s_suppkey = supplier_no
    AND total_revenue = (
        SELECT
            max(total_revenue)
        FROM (
            SELECT
                l_suppkey AS supplier_no,
                sum(l_extendedprice * (1 - l_discount)) AS total_revenue
            FROM
                lineitem
            WHERE
                l_shipdate >= CAST('1996-01-01' AS date)
                AND l_shipdate < CAST('1996-04-01' AS date)
            GROUP BY
                supplier_no) revenue1)
ORDER BY
    s_suppkey;
"""
q15_sql = q15_sql.strip().strip(";")


def q15(lineitem, supplier, dialect="duckdb", **kwargs):
    return lineitem.sql(q15_sql, dialect=dialect)


q16_sql = """
SELECT
    p_brand,
    p_type,
    p_size,
    count(DISTINCT ps_suppkey) AS supplier_cnt
FROM
    partsupp,
    part
WHERE
    p_partkey = ps_partkey
    AND p_brand <> 'Brand#45'
    AND p_type NOT LIKE 'MEDIUM POLISHED%'
    AND p_size IN (49, 14, 23, 45, 19, 3, 36, 9)
    AND ps_suppkey NOT IN (
        SELECT
            s_suppkey
        FROM
            supplier
        WHERE
            s_comment LIKE '%Customer%Complaints%')
GROUP BY
    p_brand,
    p_type,
    p_size
ORDER BY
    supplier_cnt DESC,
    p_brand,
    p_type,
    p_size;
"""
q16_sql = q16_sql.strip().strip(";")


def q16(partsupp, part, supplier, dialect="duckdb", **kwargs):
    return partsupp.sql(q16_sql, dialect=dialect)


q17_sql = """
SELECT
    sum(l_extendedprice) / 7.0 AS avg_yearly
FROM
    lineitem,
    part
WHERE
    p_partkey = l_partkey
    AND p_brand = 'Brand#23'
    AND p_container = 'MED BOX'
    AND l_quantity < (
        SELECT
            0.2 * avg(l_quantity)
        FROM
            lineitem
        WHERE
            l_partkey = p_partkey);
"""
q17_sql = q17_sql.strip().strip(";")


def q17(lineitem, part, dialect="duckdb", **kwargs):
    return lineitem.sql(q17_sql, dialect=dialect)


q18_sql = """
SELECT
    c_name,
    c_custkey,
    o_orderkey,
    o_orderdate,
    o_totalprice,
    sum(l_quantity) as sum_qty
FROM
    customer,
    orders,
    lineitem
WHERE
    o_orderkey IN (
        SELECT
            l_orderkey
        FROM
            lineitem
        GROUP BY
            l_orderkey
        HAVING
            sum(l_quantity) > 300)
    AND c_custkey = o_custkey
    AND o_orderkey = l_orderkey
GROUP BY
    c_name,
    c_custkey,
    o_orderkey,
    o_orderdate,
    o_totalprice
ORDER BY
    o_totalprice DESC,
    o_orderdate
LIMIT 100;
"""
q18_sql = q18_sql.strip().strip(";")


def q18(customer, orders, lineitem, dialect="duckdb", **kwargs):
    return customer.sql(q18_sql, dialect=dialect)


q19_sql = """
SELECT
    sum(l_extendedprice * (1 - l_discount)) AS revenue
FROM
    lineitem,
    part
WHERE (p_partkey = l_partkey
    AND p_brand = 'Brand#12'
    AND p_container IN ('SM CASE', 'SM BOX', 'SM PACK', 'SM PKG')
    AND l_quantity >= 1
    AND l_quantity <= 1 + 10
    AND p_size BETWEEN 1 AND 5
    AND l_shipmode IN ('AIR', 'AIR REG')
    AND l_shipinstruct = 'DELIVER IN PERSON')
    OR (p_partkey = l_partkey
        AND p_brand = 'Brand#23'
        AND p_container IN ('MED BAG', 'MED BOX', 'MED PKG', 'MED PACK')
        AND l_quantity >= 10
        AND l_quantity <= 10 + 10
        AND p_size BETWEEN 1 AND 10
        AND l_shipmode IN ('AIR', 'AIR REG')
        AND l_shipinstruct = 'DELIVER IN PERSON')
    OR (p_partkey = l_partkey
        AND p_brand = 'Brand#34'
        AND p_container IN ('LG CASE', 'LG BOX', 'LG PACK', 'LG PKG')
        AND l_quantity >= 20
        AND l_quantity <= 20 + 10
        AND p_size BETWEEN 1 AND 15
        AND l_shipmode IN ('AIR', 'AIR REG')
        AND l_shipinstruct = 'DELIVER IN PERSON');
"""
q19_sql = q19_sql.strip().strip(";")


def q19(lineitem, part, dialect="duckdb", **kwargs):
    return lineitem.sql(q19_sql, dialect=dialect)


q20_sql = """
SELECT
    s_name,
    s_address
FROM
    supplier,
    nation
WHERE
    s_suppkey IN (
        SELECT
            ps_suppkey
        FROM
            partsupp
        WHERE
            ps_partkey IN (
                SELECT
                    p_partkey
                FROM
                    part
                WHERE
                    p_name LIKE 'forest%')
                AND ps_availqty > (
                    SELECT
                        0.5 * sum(l_quantity)
                    FROM
                        lineitem
                    WHERE
                        l_partkey = ps_partkey
                        AND l_suppkey = ps_suppkey
                        AND l_shipdate >= CAST('1994-01-01' AS date)
                        AND l_shipdate < CAST('1995-01-01' AS date)))
            AND s_nationkey = n_nationkey
            AND n_name = 'CANADA'
        ORDER BY
            s_name;
"""
q20_sql = q20_sql.strip().strip(";")


def q20(supplier, nation, partsupp, part, lineitem, dialect="duckdb", **kwargs):
    return supplier.sql(q20_sql, dialect=dialect)


q21_sql = """
SELECT
    s_name,
    count(*) AS numwait
FROM
    supplier,
    lineitem l1,
    orders,
    nation
WHERE
    s_suppkey = l1.l_suppkey
    AND o_orderkey = l1.l_orderkey
    AND o_orderstatus = 'F'
    AND l1.l_receiptdate > l1.l_commitdate
    AND EXISTS (
        SELECT
            *
        FROM
            lineitem l2
        WHERE
            l2.l_orderkey = l1.l_orderkey
            AND l2.l_suppkey <> l1.l_suppkey)
    AND NOT EXISTS (
        SELECT
            *
        FROM
            lineitem l3
        WHERE
            l3.l_orderkey = l1.l_orderkey
            AND l3.l_suppkey <> l1.l_suppkey
            AND l3.l_receiptdate > l3.l_commitdate)
    AND s_nationkey = n_nationkey
    AND n_name = 'SAUDI ARABIA'
GROUP BY
    s_name
ORDER BY
    numwait DESC,
    s_name
LIMIT 100;
"""
q21_sql = q21_sql.strip().strip(";")


def q21(supplier, lineitem, orders, nation, dialect="duckdb", **kwargs):
    return supplier.sql(q21_sql, dialect=dialect)


q22_sql = """
SELECT
    cntrycode,
    count(*) AS numcust,
    sum(c_acctbal) AS totacctbal
FROM (
    SELECT
        substring(c_phone FROM 1 FOR 2) AS cntrycode,
        c_acctbal
    FROM
        customer
    WHERE
        substring(c_phone FROM 1 FOR 2) IN ('13', '31', '23', '29', '30', '18', '17')
        AND c_acctbal > (
            SELECT
                avg(c_acctbal)
            FROM
                customer
            WHERE
                c_acctbal > 0.00
                AND substring(c_phone FROM 1 FOR 2) IN ('13', '31', '23', '29', '30', '18', '17'))
            AND NOT EXISTS (
                SELECT
                    *
                FROM
                    orders
                WHERE
                    o_custkey = c_custkey)) AS custsale
GROUP BY
    cntrycode
ORDER BY
    cntrycode;
"""
q22_sql = q22_sql.strip().strip(";")


def q22(customer, orders, dialect="duckdb", **kwargs):
    return customer.sql(q22_sql, dialect=dialect)


all_queries = [
    q1,
    q2,
    q3,
    q4,
    q5,
    q6,
    q7,
    q8,
    q9,
    q10,
    q11,
    q12,
    q13,
    q14,
    q15,
    q16,
    q17,
    q18,
    q19,
    q20,
    q21,
    q22,
]



import os
import ibis

from ibis_bench.utils.logging import log


def get_data_dir(sf, n_partitions, csv: bool = False):
    dir_name = "csv" if csv else "parquet"
    return os.path.join("tpch_data", dir_name, f"sf={sf}", f"n={n_partitions}")


def generate_data(sf, n_partitions, csv: bool = False):
    log.info(f"generating data for sf={sf}, n={n_partitions}...")

    con = ibis.connect("duckdb://")
    con.raw_sql("PRAGMA disable_progress_bar;")

    parquet_data_directory = get_data_dir(sf, n_partitions)
    csv_data_directory = get_data_dir(sf, n_partitions, csv=True)

    if not os.path.exists(parquet_data_directory):
        for i in range(n_partitions):
            con.raw_sql(f"call dbgen(sf={sf}, children={n_partitions}, step={i})")
            for table in con.list_tables():
                if i == 0:
                    os.makedirs(
                        os.path.join(parquet_data_directory, table), exist_ok=True
                    )
                    if csv:
                        os.makedirs(
                            os.path.join(csv_data_directory, table), exist_ok=True
                        )
                log.info(
                    f"\twriting {os.path.join(parquet_data_directory, table, f'{i:04d}.parquet')}..."
                )
                con.table(table).to_parquet(
                    os.path.join(parquet_data_directory, table, f"{i:04d}.parquet")
                )
                log.info(
                    f"\tdone writing {os.path.join(parquet_data_directory, table, f'{i:04d}.parquet')}"
                )
                if csv:
                    log.info(
                        f"\twriting {os.path.join(csv_data_directory, table, f'{i:04d}.csv')}..."
                    )
                    con.table(table).to_csv(
                        os.path.join(csv_data_directory, table, f"{i:04d}.csv")
                    )
                    log.info(
                        f"\tdone writing {os.path.join(csv_data_directory, table, f'{i:04d}.csv')}"
                    )

                con.drop_table(table)
    else:
        log.info(f"\tdata already exists at {parquet_data_directory}, skipping...")


def _parquets_to_csvs():
    con = ibis.connect("duckdb://")
    for root, _, files in os.walk(os.path.join("tpch_data", "parquet")):
        for file in files:
            if file.endswith(".parquet"):
                parquet_file = os.path.join(root, file)
                csv_file = parquet_file.replace("parquet", "csv")

                os.makedirs(os.path.dirname(csv_file), exist_ok=True)

                log.info(f"reading {parquet_file}...")
                t = con.read_parquet(parquet_file)

                log.info(f"writing {csv_file}...")
                t.to_csv(csv_file)
